import json
import logging
from subprocess import check_output
from typing import Awaitable, Callable

from nats.aio.msg import Msg

from kamikaze3 import ftp
from kamikaze3.dbus.interfaces.nginx_config_reloader import NGINX_CONFIG_RELOADER
from kamikaze3.event_listener.client import create_response
from kamikaze3.event_listener.settings import PROGRAM_NAME

logger = logging.getLogger(PROGRAM_NAME)


def event_handler_wrapper(
    handler: Callable[[Msg], Awaitable]
) -> Callable[[Msg], Awaitable]:
    async def _wrapper(msg: Msg):
        logger.info(f"Received event {msg.subject}")
        resp = await handler(msg)
        logger.debug(f"Handled event {msg.subject}")
        return resp

    return _wrapper


async def reload_hmv(msg: Msg):
    logger.debug("Reloading HMV")
    check_output(
        [
            "/usr/bin/hypernode-manage-vhosts",
            "--no",
            "--all",
            "--ssl-noclobber",
            "--skip-event",
        ]
    )
    logger.debug("Done reloading HMV")


async def list_vhosts(msg: Msg):
    vhosts = check_output(
        "/usr/sbin/nginx -T 2>&1 | /usr/bin/grep -E '^\s*server_name ' | /usr/bin/awk '{print$2}' | /usr/bin/cut -d';' -f1 | /usr/bin/sort | /usr/bin/uniq",
        shell=True,
    )
    logger.debug(f"Found vhosts: {vhosts.decode()}")
    return await msg.respond(vhosts)


async def show_fpm_status(msg: Msg):
    status = check_output("/usr/bin/hypernode-fpm-status", shell=True)
    logger.debug(f"Found FPM status: {status.decode()}")
    return await msg.respond(status)


async def show_ftp_users(msg: Msg):
    try:
        users = ftp.list_users()
    except IOError:
        logger.error("IOError: Could not list FTP users")
        return await msg.respond(
            create_response(message="Could not list FTP users", status=1)
        )

    logger.debug(f"Found FTP users: {users}")
    return await msg.respond(create_response(message=None, status=0, data=users))


async def add_ftp_user(msg: Msg):
    try:
        data = json.loads(msg.data.decode())
    except json.JSONDecodeError as e:
        logger.error(f"FTP user could not be added due {str(e)}")
        return await msg.respond(
            create_response(message="Ftp user could not be added", status=1)
        )

    for required_field in ["user", "password", "homedir"]:
        if required_field in data:
            continue
        logger.error(f"FTP user could not be added. Missing {required_field}")
        return await msg.respond(
            create_response(
                message=f"FTP user could not be added. Missing {required_field}",
                status=1,
            )
        )

    try:
        user = ftp.add_user(data["user"], data["password"], data["homedir"])
    except SystemExit as e:
        logger.error(f"FTP user could not be added due {str(e)}")
        return await msg.respond(create_response(message=str(e), status=1))

    logger.info(f"FTP user added: {data['user']} for {data['homedir']}")
    return await msg.respond(create_response(message=None, status=0, data=user))


async def remove_ftp_user(msg: Msg):
    try:
        data = json.loads(msg.data.decode())
    except json.JSONDecodeError as e:
        logger.error(f"FTP user could not be removed due {str(e)}")
        return await msg.respond(
            create_response(message="Ftp user could not be removed", status=1)
        )

    if "user" not in data:
        logger.error("FTP user could not be removed. Missing username in msg")
        return await msg.respond(
            create_response(
                message="FTP user could not be removed, field user is required",
                status=1,
            )
        )

    try:
        user = ftp.remove_user(data["user"])
    except SystemExit as e:
        logger.error(f"FTP user could not be removed due {str(e)}")
        return await msg.respond(create_response(message=str(e), status=1))

    logger.info(f"FTP user removed: {data['user']}")
    return await msg.respond(create_response(message=None, status=0, data=user))


# Useful for debugging
async def pong(msg: Msg):
    logger.debug(f"Received ping from {msg.subject}")
    return await msg.respond(b"pong")


async def reload_nginx(msg: Msg):
    logger.debug("Triggering nginx reload")
    nginx_config_reloader_proxy = NGINX_CONFIG_RELOADER.get_proxy()
    nginx_config_reloader_proxy.Reload()
    return await msg.respond(b"reloaded")
