import argparse
import re
import subprocess
import sys
from os import getuid
from typing import Iterator, List, Optional, Tuple

EXPORTER_PATH = "/var/lib/prometheus/node-exporter/redis.prom"
ALLOWED_METRICS = (
    "used_memory",
    "maxmemory",
    "total_",
    "instantaneous_",
    "sync_",
    "expired_",
    "evicted_",
    "keyspace_",
    "active_defrag_",
    "connected_slaves_",
    "master_repl_",
    "repl_backlog_",
    "connected_clients_",
    "maxclients",
    "blocked_clients",
    "tracking_clients",
    "clients_in_timeout_table",
    "db[0-9]+",
)
ALLOWED_METRICS_PATTERN = re.compile("^({})".format("|".join(ALLOWED_METRICS)))
BLOCKED_METRICS_PATTERN = re.compile("(_human|_perc|_policy)$")
LSOF_TIMEOUT = 5


class TranslationDict(dict):
    def __missing__(self, key):
        return key


translations = TranslationDict(
    {
        "redis_maxclients": "redis_config_maxclients",
        "redis_used_memory": "redis_memory_used_bytes",
        "redis_used_memory_rss": "redis_memory_used_rss_bytes",
        "redis_used_memory_peak": "redis_memory_used_peak_bytes",
        "redis_used_memory_overhead": "redis_memory_used_overhead_bytes",
        "redis_used_memory_startup": "redis_memory_used_startup_bytes",
        "redis_used_memory_dataset": "redis_memory_used_dataset_bytes",
        "redis_total_system_memory": "redis_total_system_memory_bytes",
        "redis_used_memory_lua": "redis_memory_used_lua_bytes",
        "redis_used_memory_scripts": "redis_memory_used_scripts_bytes",
        "redis_maxmemory": "redis_memory_max_bytes",
        "redis_total_connections_received": "redis_connections_received_total",
        "redis_total_commands_processed": "redis_commands_processed_total",
        "redis_total_net_input_bytes": "redis_net_input_bytes_total",
        "redis_total_net_output_bytes": "redis_net_output_bytes_total",
        "redis_sync_full": "redis_replica_resyncs_full",
        "redis_sync_partial_ok": "redis_replica_partial_resync_accepted",
        "redis_sync_partial_err": "redis_replica_partial_resync_denied",
        "redis_expired_keys": "redis_expired_keys_total",
        "redis_expired_time_cap_reached_count": "redis_expired_time_cap_reached_total",
        "redis_evicted_keys": "redis_evicted_keys_total",
        "redis_keyspace_hits": "redis_keyspace_hits_total",
        "redis_keyspace_misses": "redis_keyspace_misses_total",
        "redis_active_defrag_hits": "redis_defrag_hits",
        "redis_active_defrag_misses": "redis_defrag_misses",
        "redis_active_defrag_key_hits": "redis_defrag_key_hits",
        "redis_active_defrag_key_misses": "redis_defrag_key_misses",
        "redis_repl_backlog_active": "redis_repl_backlog_is_active",
        "redis_repl_backlog_size": "redis_replication_backlog_bytes",
        "redis_repl_backlog_histlen": "redis_repl_backlog_history_bytes",
        "redis_db_expires": "redis_db_keys_expiring",
        "redis_db_avg_ttl": "redis_db_avg_ttl_seconds",
    }
)


class RedisDiscoveryError(Exception):
    pass


def translate(metric):
    metric_name = metric.split("{")[0]
    return metric.replace(metric_name, translations[metric_name])


def get_redis_instances() -> List[int]:
    output = subprocess.check_output(
        ["/usr/bin/pgrep", "-f", "redis-server"], encoding="utf-8"
    )
    output = output.strip()
    if not output:
        return []
    return [int(pid) for pid in output.split("\n")]


def get_redis_command(pid: int) -> Tuple[List[str], str]:
    try:
        lsof_output = subprocess.check_output(
            ["/usr/bin/lsof", "-P", "-a", "-p", str(pid)],
            encoding="utf-8",
            timeout=LSOF_TIMEOUT,
        )
        lsof_output = lsof_output.strip()
    except subprocess.TimeoutExpired:
        raise RedisDiscoveryError(
            f"Timed out trying to find tcp/socket information for pid: {pid}"
        )

    if not lsof_output:
        raise RedisDiscoveryError(
            f"Failed to find tcp/socket information for pid: {pid}"
        )

    port: Optional[int] = None
    socket: Optional[str] = None
    for line in lsof_output.strip().split("\n")[1:]:
        _, _, _, _, fd_type, _, _, *values = line.split()
        if fd_type == "IPv4" and values[-1].endswith("(LISTEN)"):
            port = int(values[-2].split(":")[-1])
        elif fd_type == "unix":
            socket = values[1]

    if port:
        return ["/usr/bin/redis-cli", "-p", str(port), "info"], str(port)
    elif socket:
        if not socket.startswith("/"):
            raise RedisDiscoveryError("Found redis with non-absolute socket")
        return ["/usr/bin/redis-cli", "-s", socket, "info"], socket
    else:
        raise RedisDiscoveryError(f"Failed to find port/socket for pid: {pid}")


def parse_redis_info_line(line: str, target: str) -> Iterator[str]:
    if not line or line.startswith("#") or not ALLOWED_METRICS_PATTERN.match(line):
        return []

    stat, value = [c.strip() for c in line.split(":")]
    if re.match(r"^db[0-9]+$", stat):
        # keys=258,expires=0,avg_ttl=0
        for record in value.split(","):
            db_stat, db_value = record.split("=")
            yield translate(
                'redis_db_{}{{target="{}",db="{}"}} {}'.format(
                    db_stat, target, stat, db_value
                )
            )
    elif not BLOCKED_METRICS_PATTERN.search(stat):
        yield translate('redis_{}{{target="{}"}} {}'.format(stat, target, value))

    return []


def get_redis_info(pid) -> List[str]:
    result = []
    try:
        command, target = get_redis_command(pid)
        output = subprocess.check_output(command, encoding="utf-8")
        for line in output.strip().split("\n"):
            for parsed in parse_redis_info_line(line, target):
                result.append(parsed)
    except RedisDiscoveryError as e:
        print(f"ERROR: {e}", file=sys.stderr)

    return result


def parse_args(args: Optional[List[str]] = None) -> bool:
    parser = argparse.ArgumentParser(
        description="Collect and write redis information for prometheus node exporter"
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        default=False,
        help="Enable verbose logging",
    )

    parsed_args = parser.parse_args(args)
    return parsed_args.verbose


def main():
    verbose = parse_args()

    if getuid() != 0:
        print("ERROR: This command must be run as root.", file=sys.stderr)
        sys.exit(1)

    pids = get_redis_instances()
    result = []

    for pid in pids:
        result.extend(get_redis_info(pid))

    with open(EXPORTER_PATH, "w", encoding="utf-8") as f:
        for metric in result:
            if verbose:
                print(metric)
            f.write(metric + "\n")


if __name__ == "__main__":
    main()
